/*
 * aic3106-conf.c
 *
 * TLV320AIC3106 codec configurator
 *
 * Copyright (C) 2016 Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * Based on sound/soc/codecs/tlv320aic3x.c by Vladimir Barinov
 * Based on sound/soc/codecs/wm8753.c by Liam Girdwood
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <linux/input.h>
#ifdef HAVE_LINUX_I2C_DEV_H
#include <linux/i2c-dev.h>
#else
#include "i2c-dev.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <tlv320aic3x.h>

typedef unsigned char u8;

struct reg_default {
	unsigned int reg;
	unsigned int def;
};

struct codec {
	int fd;
	unsigned int addr;
	unsigned char *regs_cache;
	unsigned int num_regs;
};

/* I2C bus device. This requires the i2c-dev kernel module. */
static const char i2c_device[] = "/dev/i2c-1";

static void *xmalloc(size_t size)
{
	void *p = malloc(size);

	if (!p) {
		fprintf(stderr, "error: can't allocate memory\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

static const struct reg_default aic3x_reg[] = {
	{   0, 0x00 }, {   1, 0x00 }, {   2, 0x00 }, {   3, 0x10 },
	{   4, 0x04 }, {   5, 0x00 }, {   6, 0x00 }, {   7, 0x00 },
	{   8, 0x00 }, {   9, 0x00 }, {  10, 0x00 }, {  11, 0x01 },
	{  12, 0x00 }, {  13, 0x00 }, {  14, 0x00 }, {  15, 0x80 },
	{  16, 0x80 }, {  17, 0xff }, {  18, 0xff }, {  19, 0x78 },
	{  20, 0x78 }, {  21, 0x78 }, {  22, 0x78 }, {  23, 0x78 },
	{  24, 0x78 }, {  25, 0x00 }, {  26, 0x00 }, {  27, 0xfe },
	{  28, 0x00 }, {  29, 0x00 }, {  30, 0xfe }, {  31, 0x00 },
	{  32, 0x18 }, {  33, 0x18 }, {  34, 0x00 }, {  35, 0x00 },
	{  36, 0x00 }, {  37, 0x00 }, {  38, 0x00 }, {  39, 0x00 },
	{  40, 0x00 }, {  41, 0x00 }, {  42, 0x00 }, {  43, 0x80 },
	{  44, 0x80 }, {  45, 0x00 }, {  46, 0x00 }, {  47, 0x00 },
	{  48, 0x00 }, {  49, 0x00 }, {  50, 0x00 }, {  51, 0x04 },
	{  52, 0x00 }, {  53, 0x00 }, {  54, 0x00 }, {  55, 0x00 },
	{  56, 0x00 }, {  57, 0x00 }, {  58, 0x04 }, {  59, 0x00 },
	{  60, 0x00 }, {  61, 0x00 }, {  62, 0x00 }, {  63, 0x00 },
	{  64, 0x00 }, {  65, 0x04 }, {  66, 0x00 }, {  67, 0x00 },
	{  68, 0x00 }, {  69, 0x00 }, {  70, 0x00 }, {  71, 0x00 },
	{  72, 0x04 }, {  73, 0x00 }, {  74, 0x00 }, {  75, 0x00 },
	{  76, 0x00 }, {  77, 0x00 }, {  78, 0x00 }, {  79, 0x00 },
	{  80, 0x00 }, {  81, 0x00 }, {  82, 0x00 }, {  83, 0x00 },
	{  84, 0x00 }, {  85, 0x00 }, {  86, 0x00 }, {  87, 0x00 },
	{  88, 0x00 }, {  89, 0x00 }, {  90, 0x00 }, {  91, 0x00 },
	{  92, 0x00 }, {  93, 0x00 }, {  94, 0x00 }, {  95, 0x00 },
	{  96, 0x00 }, {  97, 0x00 }, {  98, 0x00 }, {  99, 0x00 },
	{ 100, 0x00 }, { 101, 0x00 }, { 102, 0x02 }, { 103, 0x00 },
	{ 104, 0x00 }, { 105, 0x00 }, { 106, 0x00 }, { 107, 0x00 },
	{ 108, 0x00 }, { 109, 0x00 },
};

static int aic3x_write(int fd, unsigned int addr, unsigned int x,
		unsigned int reg)
{
	struct i2c_msg xfer[1];
	struct i2c_rdwr_ioctl_data rdwr;
	unsigned char buf[2];
	int ret;

	buf[0] = reg;
	buf[1] = x & 0xff;

	memset(xfer, 0, sizeof(xfer));
	xfer[0].addr = addr;
	xfer[0].flags = 0;
	xfer[0].len = sizeof(buf);
	xfer[0].buf = (void *) buf;

	memset(&rdwr, 0, sizeof(rdwr));
	rdwr.msgs = xfer;
	rdwr.nmsgs = 1;

	ret = ioctl(fd, I2C_RDWR, &rdwr);
	if (ret < 0) {
		fprintf(stderr, "%02x: 0x%03x -> %02x: %s\n", addr, x, reg,
				strerror(errno));
		return -errno;
	}
	return 0;
}

static unsigned int snd_soc_read(struct codec *codec, unsigned int reg)
{
	if (reg >= codec->num_regs)
		return -1;

	return codec->regs_cache[reg];
}

static int snd_soc_write(struct codec *codec, unsigned int reg,
		unsigned int val)
{
	if (reg >= codec->num_regs)
		return -EINVAL;

	codec->regs_cache[reg] = val;
	aic3x_write(codec->fd, codec->addr, val, reg);
	return 0;
}

static int snd_soc_update_bits(struct codec *codec, unsigned int reg,
				unsigned int mask, unsigned int val)
{
	unsigned char orig, tmp;

	if (reg > codec->num_regs)
		return -EINVAL;

	orig = codec->regs_cache[reg];
	tmp = orig & ~mask;
	tmp |= val & mask;
	if (tmp != orig)
		return snd_soc_write(codec, reg, tmp);
	return 0;
}

int main(int argc, char **argv)
{
	struct codec *codec;
	int fd;
	int x;

	fd = open(i2c_device, O_RDWR);
	if (fd < 0) {
		perror("opening i2c device");
		return EXIT_FAILURE;
	}

	codec = xmalloc(sizeof(*codec));
	memset(codec, 0, sizeof(*codec));
	codec->fd = fd;
	codec->addr = 0x18;
	codec->num_regs = DAC_ICC_ADJ + 1;
	codec->regs_cache = xmalloc(codec->num_regs);

	snd_soc_write(codec, AIC3X_PAGE_SELECT, PAGE0_SELECT);
	x = snd_soc_read(codec, AIC3X_PAGE_SELECT);
	if (x < 0) {
		fprintf(stderr, "error: page select read error\n");
		exit(EXIT_FAILURE);
	}
	if (x & 1) {
		fprintf(stderr, "error: page select read error: %02x\n", x);
		exit(EXIT_FAILURE);
	}
	snd_soc_write(codec, AIC3X_RESET, SOFT_RESET);

	/* DAC default volume and mute */
	snd_soc_write(codec, LDAC_VOL, DEFAULT_VOL | MUTE_ON);
	snd_soc_write(codec, RDAC_VOL, DEFAULT_VOL | MUTE_ON);
	snd_soc_write(codec, DAC_PWR, 0xe0);
	snd_soc_write(codec, HPRCOM_CFG, 0x16);
	snd_soc_write(codec, DAC_LINE_MUX, 0x50);

	snd_soc_write(codec, AIC3X_HEADSET_DETECT_CTRL_B, 0x80);

	/* DAC to HP default volume and route to Output mixer */
	snd_soc_write(codec, DACL1_2_HPLOUT_VOL, DEFAULT_VOL | ROUTE_ON);
	snd_soc_write(codec, DACR1_2_HPROUT_VOL, DEFAULT_VOL | ROUTE_ON);
	snd_soc_write(codec, DACL1_2_HPLCOM_VOL, DEFAULT_VOL | ROUTE_ON);
	snd_soc_write(codec, DACR1_2_HPRCOM_VOL, DEFAULT_VOL | ROUTE_ON);

	/* unmute all outputs */
	snd_soc_update_bits(codec, LLOPM_CTRL, UNMUTE, UNMUTE);
	snd_soc_update_bits(codec, RLOPM_CTRL, UNMUTE, UNMUTE);
	snd_soc_update_bits(codec, HPLOUT_CTRL, UNMUTE, UNMUTE);
	snd_soc_update_bits(codec, HPROUT_CTRL, UNMUTE, UNMUTE);
	snd_soc_update_bits(codec, HPLCOM_CTRL, UNMUTE, UNMUTE);
	snd_soc_update_bits(codec, HPRCOM_CTRL, UNMUTE, UNMUTE);

	/* ADC default volume and unmute */
	snd_soc_write(codec, LADC_VOL, 0x00);
	snd_soc_write(codec, RADC_VOL, 0x00);
	/* By default route Line1 to ADC PGA mixer */
	snd_soc_write(codec, LINE1L_2_LADC_CTRL, 0x4);
	snd_soc_write(codec, LINE1R_2_RADC_CTRL, 0x4);
	snd_soc_write(codec, MIC3LR_2_LADC_CTRL, 0xf0);
	snd_soc_write(codec, MIC3LR_2_RADC_CTRL, 0x0f);
	snd_soc_write(codec, MICBIAS_CTRL, 0xc0);

	/* PGA to HP Bypass default volume, disconnect from Output Mixer */
	snd_soc_write(codec, PGAL_2_HPLOUT_VOL, DEFAULT_VOL);
	snd_soc_write(codec, PGAR_2_HPROUT_VOL, DEFAULT_VOL);
	snd_soc_write(codec, PGAL_2_HPLCOM_VOL, DEFAULT_VOL);
	snd_soc_write(codec, PGAR_2_HPRCOM_VOL, DEFAULT_VOL);
	/* PGA to Line Out default volume, disconnect from Output Mixer */
	snd_soc_write(codec, PGAL_2_LLOPM_VOL, DEFAULT_VOL);
	snd_soc_write(codec, PGAR_2_RLOPM_VOL, DEFAULT_VOL);

	snd_soc_write(codec, AIC3X_CODEC_DATAPATH_REG,
			LDAC2LCH | RDAC2RCH | FSREF_48000 | DUAL_RATE_MODE);
	snd_soc_write(codec, AIC3X_SAMPLE_RATE_SEL_REG, 0);

	/*
	 * configure samplerate, 25.576 MHz clock,
	 * divided by (4 * 128) = 48 kHz
	 */
	snd_soc_write(codec, AIC3X_PLL_PROGA_REG, 4 << PLLQ_SHIFT);
	snd_soc_write(codec, AIC3X_GPIOB_REG, CODEC_CLKIN_CLKDIV);

	/*
	 * configure interface, 24-bit I2S in Master mode
	 */
	snd_soc_write(codec, AIC3X_ASD_INTF_CTRLA,
			BIT_CLK_MASTER | WORD_CLK_MASTER | DOUT_TRISTATE);
	snd_soc_write(codec, AIC3X_ASD_INTF_CTRLB,
			AUDIO_WORD_LENGTH_32);

	snd_soc_write(codec, LDAC_VOL, 0x0);
	snd_soc_write(codec, RDAC_VOL, 0x0);

	return 0;
}
